classdef AdaptiveLocalLearning
    
    % ---------------------------------------------------------------------
    % Supported functionalities:
    %   
    %   1. Learning hybrid templates category by category with three types of features.
    %
    %   2. Cross-validation and testing category by category.
    %
    % 
    % Future work:
    %
    %   1. Use a standardized image pyramid and a better modularized
    %       convolution method.
    %
    % 
    % ---------------------------------------------------------------------
    
    % member variables
    properties
        categories;	% collection of category names
        iRun;		% index for cross validation run
        iCategory;	% index of current category
        category;	% name of current category
        OverwriteModeIsOn = true;
    end

    properties
        maxImArea = 1000*1000;	% maximum allowed image area
        posFolder;		% folder for positive example images
        negFolder = 'Image/Natural';	% folder for random natural images
        featureVersion = 'GaborS1';		% feature name
        S1FeatureType = 'S1LocalNorm';
        windowSize = 120; %
        Totalsketch = 32; % maximum allowed number of features
        epsilon = .1; % allowed correlation between selected Gabors
        sub = 1;    % subsampling rate for Gabor
        Lrange = 6; % allowed shifting in location (times sub pixels)
        Orange = 1; % allowed shifting in orientation
        norient = 16; % number of Gabor orientations
        %%% Note: Lrange*sub is the pixels allowed to shift along normal
        %%%       Lrange*sub should be the same in learning and testing
        SHUTUP = 0.; % SHUTUP is to neglect the positions with low average responses
        ToFindWhileLearning = true; % use EM-type iterations for learning templates for objects with unknown locations
        sizeGabor = 17;
        gainThres = 0.1;
		test_gainThres = 0.2;
		discount_S2 = 1;
        max_EM_iter = 20;
        displayModeIsOn = false;
        saturation = 6;
    end
    
    properties
        ap;				% average precision
        testAUC;		% area under ROC curve
        tp;				% true positive rates
        fp;				% false positive rates
        rec;			% recall rates
        prec;			% precision rates
        EER;			% equal-error rate
    end
    
    % functions
    methods
        
        % constructor
        function obj = SketchTextureFlat(iRun, iCat, categories)
            obj.iRun = iRun;
            obj.iCategory = iCat;
            obj.categories = categories;
            obj.category = categories{iCat};
        end
        
        

        %% learns multiple templates from a set of images (positive examples)
        %{ 
        The procedure is adaptive to the neighboring categories:
            for each image that is not explained:
                1. learn a template from single example
                2. use this template to scan all training images for all categories to get MAX2 scores
                3. select a score threshold, such that False-positive-rate <= a preset value
                4. re-learn the template using un-explained positive examples above this threshold
            
        %}
        function learn(this)

            disp(sprintf('start learning: %d %s run %d',this.iCategory,this.category,this.iRun));
            close all
            if this.displayModeIsOn
                figure;
                imshow( zeros(this.windowSize,this.windowSize),[] );
            end
            
			load('Feature/naturalStat_GaborS1_LocalNorm.mat'); % load natural statistics
            % TRAINING EXAMPLES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % ==================================================================
            %  input images from a folder
            % ==================================================================
            file0 = sprintf('Split/%s/split%d.mat',this.category,this.iRun);
            load(file0,'posTrainImages');
            nimage = length(posTrainImages);
            
            % prepare Gabor filters and compute the inhibition map
            tic
            [allfilter, allsymbol] = makefilter(scale, this.norient);  % generate Gabor filters
            for o = 1:this.norient
                allsymbol{o} = single(allsymbol{o});
            end
            h = (size(allfilter{1}, 1)-1)/2;  % half size of Gabor
            C = corr(allfilter, this.epsilon);    % generate the inhibition maps
            for o = 1:this.norient
                for o2 = 1:this.norient
                    C{o,o2} = single(C{o,o2});
                end
            end
            
            % load and inscribe images into the template lattice
            SUM1mapLearn = cell(nimage,this.norient);
            stored_SUM1map = cell(nimage,this.norient);
            for i = 1:nimage
                load(sprintf('Feature/%s/%s/%s.mat',...
                    this.featureVersion,this.category,posTrainImages{i}),this.S1FeatureType);
                % inscribe the S1 map into a square bounding box
                for o = 1:this.norient
                    SUM1mapLearn{i,o} = zeros(this.windowSize, this.windowSize,'single');
                    tmp = inscribe([this.windowSize this.windowSize],S1LocalNorm{o},0);
                    SUM1mapLearn{i,o}(:) = tmp(:);
                end
            end
			
            for i = 1:nimage
            	for o = 1:this.norient
					stored_SUM1map{i,o} = zeros(size(SUM1mapLearn{i,o}),'single');
					stored_SUM1map{i,o}(:) = SUM1mapLearn{i,o}(:);
            	end
            end
            
            % -----------------------------------------------
            % begin EM-type iterations:
            % -----------------------------------------------
			if this.ToFindWhileLearning
				max_iter = this.max_EM_iter;
			else
				max_iter = 1;
            end
			
            diff_in_detection = 1; % the proportion that the current detection is different from last time
            current_row_translation = zeros(nimage,1);
            current_col_translation = zeros(nimage,1);
			for iter = 1:max_iter
			
				disp( sprintf('begin iteration %d', iter) );
				% ---- M step ----
				%
				% learning of sketch
				%
				tic
				load('Feature/naturalStat_GaborS1_LocalNorm.mat'); % load natural statistics
		        Mi = zeros(1, this.Totalsketch, 'single');  % orientation of selected Gabor
		        Mx = zeros(1, this.Totalsketch, 'single');  % position of selected Gabor
		        My = zeros(1, this.Totalsketch, 'single'); 
		        Mm = zeros(1, this.Totalsketch, 'single');  % lambda
		        Mm1 = zeros(1, this.Totalsketch, 'single');  % logZ
		        gain = zeros(1, this.Totalsketch, 'single'); % information gains (bits)
		        sym = zeros(this.windowSize,this.windowSize, 'single');  % symbolic plot of selected Gabor
		        Asym = cell(1, nimage);   % symbolic plot for each image with active Gabor
		        for i = 1 : nimage
		            Asym{i} = zeros(size(SUM1mapLearn{i,1}),'single');  
		        end
		        tss = zeros(nimage, 1, 'single');  % total scores for testing images
		        gain = zeros(this.Totalsketch, 1, 'single');  % coding gain for selected Gabor
		        binnum = 50; % number of bins
		        subsampleM1 = 1;
		        MAX1mapLearn = cell(size(SUM1mapLearn));
  				ARGMAX1mapLearn = cell(size(SUM1mapLearn));
		        for i = 1:nimage
				    [MAX1mapLearn(i,:) ARGMAX1mapLearn(i,:) M1RowShift M1ColShift M1OriShifted] = ...
				    	mexc_ComputeMAX1( this.norient, SUM1mapLearn(i,:), single(this.Lrange*this.sub)/(h*2+1),...
		            		this.Orange, single(h*2+1), subsampleM1 );
		        end

		        mexc_SharedSketch(this.norient, this.Lrange * this.sub, this.Orange, subsampleM1, ... % about active basis  
				   this.Totalsketch, nimage, this.windowSize, this.windowSize, ...
				   SUM1mapLearn, MAX1mapLearn, ARGMAX1mapLearn, ... % about training images 
				   h, C, allsymbol(1, :), ... % about filters
				   single(M), single(lam), single(e), single(lz), ... % about exponential model 
				   Mi, Mx, My, Mm, Mm1,gain, ... % learned parameters
				   sym, Asym,... % learned templates 
				   M1RowShift, M1ColShift, M1OriShifted);
		        
		        nF = max(1,sum(gain>this.gainThres));
		        Mi = Mi(1:nF);
		        Mx = Mx(1:nF);
		        My = My(1:nF);
		        Mm = Mm(1:nF);
		        Mm1 = Mm1(1:nF);
		        sym = displayGaborTemplate([this.windowSize this.windowSize], Mx,My,Mi,Mm);
		        
		        if ~exist(sprintf('Result/%s/AdaptiveLocalLearning/',this.category),'dir')
		            mkdir(sprintf('Result/%s/AdaptiveLocalLearning/',this.category));
		        end
		        
		        Lrange = this.Lrange; Orange = this.Orange; sub = this.sub;
		        S2Template = struct( 'selectedRow', single(Mx)-1-floor(this.windowSize/2), ...
		        	'selectedCol', single(My)-1-floor(this.windowSize/2), ...
		        	'selectedOri', single(Mi), ...
		        	'selectedScale', zeros(1,length(Mx),'single'), ... 
		        	'selectedLambda', single(Mm), ...
		        	'selectedLogZ', single(Mm1), ...
		        	'subsampleM1', 1, ...
					'selectedGain', gain,...
		        	'sym', sym ...
		        	);
		        
		        save(sprintf('Result/%s/AdaptiveLocalLearning/model_run%d.mat',this.category,this.iRun),...
		            'Mi','Mx','My', 'Mm','Mm1','C','h','scale',...
		            'norient','allfilter','allsymbol','Asym','Upperbound',...
		            'M','lam','e', 'lz', 'sym','tss','gain','Lrange','Orange','sub','S2Template');
		        disp(['learning time for sketch: ' num2str(toc) ' seconds']);
				
		        %% ---- E step ----
		        % prepare [SUM1mapLearn] for the next learning step
		        if iter >= max_iter
		        	break;
		        end

				% - obtain transformed version of templates
				% do nothing; (no rotation, no scaling, only translation)
				
				for iImage = 1:size(SUM1mapLearn,1)
				
					% - compute MAX1 maps
					locationPerturbationFraction = this.Lrange * sub / double(this.sizeGabor);
					orientationPerturbationRadius = this.Orange;
					subsampleM1 = 1;
					M1Map = mexc_ComputeMAX1( this.norient, stored_SUM1map(iImage,:), locationPerturbationFraction, ...
						orientationPerturbationRadius, this.sizeGabor, subsampleM1 );
				
					% - compute SUM2 maps of transformed templates on original feature maps		
					subsampleS2 = 1;
					
					S2Map = mexc_ComputeSUM2( this.norient, M1Map, {S2Template}, subsampleS2 );
					
					

					% - find the best match (global max)																						
					[Max2 bestInd] = max(S2Map{1}(:));
                    

					col_translation = ceil( bestInd / this.windowSize );
					row_translation = bestInd - (col_translation-1) * this.windowSize;
					col_translation = col_translation - 1;
					row_translation = row_translation - 1;

					% - crop SUM1 (sigmoid saturated) maps	
					denseX = -floor(this.windowSize/2) + (1:this.windowSize)-1;
					denseY = denseX;
					count = 0;
					inRow = zeros(length(denseX)*length(denseY),1,'single');
					inCol = zeros(length(denseX)*length(denseY),1,'single');
					inOri = zeros(length(denseX)*length(denseY),1,'single');
					inScale = zeros(length(denseX)*length(denseY),1,'single');
					for y = denseY
						for x = denseX
						    count = count+1;
						    inRow(count) = x;
						    inCol(count) = y;
						end
					end

					% use mex-C code to transform the dense grid, which is used to fetch SUM1 map pixels
					[outX outY] = mexc_TemplateAffineTransform( 0, 2^0, 2^0, 0, inRow, inCol, inOri, inScale, this.norient );
					
					tScale = 0; destHeight = this.windowSize; destWidth = this.windowSize; nScale = 1;
					SUM1mapLearn(iImage,:) = mexc_CropInstance( stored_SUM1map(iImage,:),...
				    	row_translation,...
				    	col_translation,...
				    	0,tScale,outX,outY,destWidth,destHeight,this.norient,1 );
                    
                    current_row_translation(iImage) = row_translation;
                    current_col_translation(iImage) = col_translation;
                    
                    
                end
                
                % compute the difference in detection
                nImage = size(SUM1mapLearn,1);
                if iter == 1
                    diff_in_detection = 1;
                else
                    count_dif = 0;
                    for iImage = 1:nImage
                        if abs( last_row_translation(iImage) - current_row_translation(iImage) ) > 3 ...
                            || abs( last_col_translation(iImage) - current_col_translation(iImage)) > 3
                            count_dif = count_dif + 1;
                        end
                    end
                    diff_in_detection = count_dif / nImage;
                end
                disp(sprintf('difference in detection: %.3f',diff_in_detection));
                last_row_translation = current_row_translation;
                last_col_translation = current_col_translation;
                
                if diff_in_detection < 0.2
                    break;
                end
                
            end

	        S2Template.extraLogZ = 0;			
			save(sprintf('Result/%s/AdaptiveLocalLearning/model_run%d.mat',this.category,this.iRun),...
		            '-append','S2Template');
        end

        % -----------------------------------------------------
		%% test on training data
		% -----------------------------------------------------

		function test_on_train_multiclass(this)
        	nT = length(this.categories);
        	
        	% go over categories to fetch templates
        	list_of_S2Template = cell(nT,1);
        	for iT = 1:nT
        		category = this.categories{iT};
        		load(sprintf('Result/%s/AdaptiveLocalLearning/model_run%d.mat',category,this.iRun),'S2Template','h','S2TemplateTexture');
				ind = find(S2Template.selectedGain > this.test_gainThres);
				S2Template.selectedRow = S2Template.selectedRow(ind);
				S2Template.selectedCol = S2Template.selectedCol(ind);
				S2Template.selectedOri = S2Template.selectedOri(ind);
				S2Template.selectedScale = S2Template.selectedScale(ind);
				S2Template.selectedLambda = S2Template.selectedLambda(ind);
				S2Template.selectedLogZ = S2Template.selectedLogZ(ind);
        		list_of_S2Template{iT} = S2Template;
				
				ind = find(S2TemplateTexture.selectedGain > this.test_gainThres);
				S2TemplateTexture.selectedRow = S2TemplateTexture.selectedRow(ind);
				S2TemplateTexture.selectedCol = S2TemplateTexture.selectedCol(ind);
				S2TemplateTexture.selectedOri = S2TemplateTexture.selectedOri(ind);
				S2TemplateTexture.selectedScale = S2TemplateTexture.selectedScale(ind);
				S2TemplateTexture.selectedLambda = S2TemplateTexture.selectedLambda(ind);
				S2TemplateTexture.selectedLogZ = S2TemplateTexture.selectedLogZ(ind);
        		list_of_S2TemplateTexture{iT} = S2TemplateTexture;
        	end
        	disp('templates loaded');

        	% Go over categories to fetch images and compute SUM2 maps
        	extraLogZ_pos = zeros( nT, 1 );
        	count_image = 0;
        	for iCategory = 1:nT
        		fprintf(1,'testing on %s: ',this.categories{iCategory});
        		tic
		    	file0 = sprintf('Split/%s/split%d.mat',this.categories{iCategory},this.iRun);
		        load(file0,'posTrainImages');
		        nimage = length(posTrainImages);
		        count_image = count_image + nimage;
		        scores_train = cell(nimage,1);
		        
		        for i = 1:nimage
		        	for kk = 1:length(num2str(i-1))
		        		if i == 1
		        		else
		        			fprintf('\b');
		        		end
		        	end
		        	fprintf(1,'%d',i);
				    % for each image, test all templates on it
				    load(sprintf('Feature/%s/%s/%s.mat',...
		                this.featureVersion,this.categories{iCategory},posTrainImages{i}),this.S1FeatureType,'S1LocalMean');

		            S1 = S1LocalNorm;
		            clear S1LocalNorm;
		            locationPerturbationFraction = this.Lrange * this.sub / double(this.sizeGabor);
					orientationPerturbationRadius = this.Orange;
					subsampleM1 = 1;
		            M1 = mexc_ComputeMAX1(this.norient, S1, locationPerturbationFraction, ...
							orientationPerturbationRadius, this.sizeGabor, subsampleM1);

				    % it is an image with certain object, then inscribe it
				    
		            % positive image patches may be smaller than the template lattice   
		            % need to be inscribed in a square bounding box
		            for o = 1:this.norient
		                M1{o} = single( inscribe([this.windowSize,this.windowSize], M1{o}, 0) );
		            end
		            
		            % - compute SUM2 maps
		            if ~exist( sprintf('Feature/GaborS2/%s',this.categories{iCategory}), 'dir' )
						mkdir( sprintf('Feature/GaborS2/%s',this.categories{iCategory}) );
					end
		            S2MapPath = sprintf('%s/%s/%s_run%d.mat','Feature/GaborS2/',this.categories{iCategory},posTrainImages{i},this.iRun);
					
					if ~exist( S2MapPath, 'file' ) || this.OverwriteModeIsOn
						subsampleS2 = 5;
						S2Map = mexc_ComputeSUM2( this.norient, M1, list_of_S2Template, subsampleS2 );
						%S2MapTexture = mexc_ComputeSUM2( this.norient, S1LocalMean, list_of_S2Template, subsampleS2 );
						
						%for ii = 1:numel(S2Map)
						%	S2Map{ii} = S2Map{ii} + S2MapTexture{ii};
						%	S2Map{ii} = this.discount_S2 * S2Map{ii};
						%end
						
						save(S2MapPath, 'S2Map');
					else % if the S2 map is already computed, then load it
						load(S2MapPath, 'S2Map');
					end
				    
				    tmp = zeros(1,nT,'single');
				    for jT = 1:nT
				    	tmp(jT) = max(S2Map{jT}(:));
                    end
                    scores_train{i} = tmp;
                    
					%for jT = 1:nT
				    %	extraLogZ_pos(jT) = extraLogZ_pos(jT) + exp( scores_train{i}(jT) );
				    %end
                end
                
		        
		        save(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCategory}, this.iRun),'-append','scores_train');
		        fprintf(1,' images; takes %.1f sec\n',toc);
            end
            
            % re-normalization using the positive training images (instead of natural images)
            %for jT = 1:nT
            %	extraLogZ_pos(jT) = log(extraLogZ_pos(jT)/count_image);
            %end
			extraLogZ_pos = zeros(1,nT); % cancel the use of extra normalization
            
            mixing = ones(1,nT) / nT;
			dif = 1;
			iIter = 1;
			
			last_voting = cell(1,nT);
			for iT = 1:nT
				last_voting{iT} = ones(1,nT) / nT;
			end

			while dif > 1e-6 && iIter < 2
				% now obtain voting statistics
				% voting{jT}{iCategory} means the conditional probability (not joint!) p( label | cluster number )
				voting = cell(1,nT);
				for iT = 1:nT
					voting{iT} = zeros( 1, nT );
				end

				post_t_given_image = zeros(1,nT);
				new_mixing = zeros(1,nT) / nT;
				for iCategory = 1:nT
					file0 = sprintf('Split/%s/split%d.mat',this.categories{iCategory},this.iRun);
					load(file0,'posTrainImages');
					nimage = length(posTrainImages);
					load(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCategory}, this.iRun),'scores_train');

					for ii = 1:nimage
						for jT = 1:nT
							% post_t_given_image(jT) = exp( scores_train{ii}(jT) - extraLogZ_pos(jT) ) * last_voting{jT}(iCategory) * mixing(jT);
							post_t_given_image(jT) = exp( scores_train{ii}(jT) - extraLogZ_pos(jT) ) * mixing(jT);
						end
						post_t_given_image = post_t_given_image / sum(post_t_given_image);
						new_mixing = new_mixing + post_t_given_image;
						
						for jT = 1:nT
							voting{jT}(iCategory) = voting{jT}(iCategory) + post_t_given_image(jT);
						end
					end
				end

				new_mixing = new_mixing / count_image;
				dif = mean(abs(new_mixing-mixing));
				mixing = new_mixing;
				last_voting = voting;
				iIter = iIter + 1;
			end

			for iT = 1:nT
				voting{iT} = voting{iT} / sum(voting{iT});
			end

            mkdir('Result/summary/GaborS1');
		    save(sprintf('Result/%s/GaborS1/voting_run%d.mat', 'summary', this.iRun),'voting','extraLogZ_pos','mixing');
        end

		%
		%% testing
        %
        function test_multiclass(this)
        	nT = length(this.categories);

        	% go over categories to fetch templates
        	list_of_S2Template = cell(nT,1);
        	for iT = 1:nT
        		category = this.categories{iT};
        		load(sprintf('Result/%s/AdaptiveLocalLearning/model_run%d.mat',category,this.iRun),'S2Template','S2TemplateTexture','h');
        		ind = find(S2Template.selectedGain > this.test_gainThres);
				S2Template.selectedRow = S2Template.selectedRow(ind);
				S2Template.selectedCol = S2Template.selectedCol(ind);
				S2Template.selectedOri = S2Template.selectedOri(ind);
				S2Template.selectedScale = S2Template.selectedScale(ind);
				S2Template.selectedLambda = S2Template.selectedLambda(ind);
				S2Template.selectedLogZ = S2Template.selectedLogZ(ind);
        		list_of_S2Template{iT} = S2Template;
				
				ind = find(S2TemplateTexture.selectedGain > this.test_gainThres);
				S2TemplateTexture.selectedRow = S2TemplateTexture.selectedRow(ind);
				S2TemplateTexture.selectedCol = S2TemplateTexture.selectedCol(ind);
				S2TemplateTexture.selectedOri = S2TemplateTexture.selectedOri(ind);
				S2TemplateTexture.selectedScale = S2TemplateTexture.selectedScale(ind);
				S2TemplateTexture.selectedLambda = S2TemplateTexture.selectedLambda(ind);
				S2TemplateTexture.selectedLogZ = S2TemplateTexture.selectedLogZ(ind);
        		list_of_S2TemplateTexture{iT} = S2TemplateTexture;
        	end
        	disp('templates loaded');

        	% Go over categories to fetch images and compute SUM2 maps
        	for iCategory = 1:nT
        		fprintf(1,'testing on %s: ',this.categories{iCategory});
        		tic
		    	file0 = sprintf('Split/%s/split%d.mat',this.categories{iCategory},this.iRun);
		        load(file0,'posTestImages');
		        nimage = length(posTestImages);
		        scores = cell(nimage,1);

		        for i = 1:nimage
		        	for kk = 1:length(num2str(i-1))
		        		if i == 1
		        		else
		        			fprintf('\b');
		        		end
		        	end
		        	fprintf(1,'%d',i);
				    % for each image, test all templates on it
				    load(sprintf('Feature/%s/%s/%s.mat',...
		                this.featureVersion,this.categories{iCategory},posTestImages{i}),this.S1FeatureType,...
                        'S1LocalNorm','S1LocalMean');

		            S1 = S1LocalNorm;
		            clear S1LocalNorm;
		            locationPerturbationFraction = this.Lrange * this.sub / double(this.sizeGabor);
					orientationPerturbationRadius = this.Orange;
					subsampleM1 = 1;
		            M1 = mexc_ComputeMAX1(this.norient, S1, locationPerturbationFraction, ...
							orientationPerturbationRadius, this.sizeGabor, subsampleM1);

				    % it is an image with certain object, then inscribe it

		            % positive image patches may be smaller than the template lattice   
		            % need to be inscribed in a square bounding box
                    %{
		            for o = 1:this.norient
		                M1{o} = single( inscribe([this.windowSize,this.windowSize], M1{o}, 0) );
		            end
                    %}

		            % - compute SUM2 maps
		            if ~exist( sprintf('Feature/GaborS2/%s',this.categories{iCategory}), 'dir' )
						mkdir( sprintf('Feature/GaborS2/%s',this.categories{iCategory}) );
					end
		            S2MapPath = sprintf('%s/%s/%s_run%d.mat','Feature/GaborS2/',this.categories{iCategory},posTestImages{i},this.iRun);

					if ~exist( S2MapPath, 'file' ) || this.OverwriteModeIsOn
						subsampleS2 = 5;
						S2Map = mexc_ComputeSUM2( this.norient, M1, list_of_S2Template, subsampleS2 );
						S2MapTexture = mexc_ComputeSUM2( this.norient, S1LocalMean, list_of_S2TemplateTexture, subsampleS2 );
						for ii = 1:numel(S2Map)
							% S2Map{ii} = S2Map{ii} + S2MapTexture{ii};
						%	S2Map{ii} = S2Map{ii} - list_of_S2Template{ii}.extraLogZ;
							S2Map{ii} = this.discount_S2 * S2Map{ii};
						end
						save(S2MapPath, 'S2Map');
					else % if the S2 map is already computed, then load it
						load(S2MapPath, 'S2Map');
					end

				    tmp = zeros(1,nT,'single');
				    for jT = 1:nT
				    	tmp(jT) = max(S2Map{jT}(:));
                    end
                    scores{i} = tmp;
		        end

		        save(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCategory}, this.iRun),'scores');
		        fprintf(1,' images; takes %.1f sec\n',toc);
            end
        end

        %%% post processing (Bayesian voting)

        function post_process(this)
        	load(sprintf('Result/%s/GaborS1/voting_run%d.mat','summary',this.iRun), 'voting', 'extraLogZ_pos', 'mixing' );
        	nCat = length(this.categories);
        	nT = nCat; % in future, nT may be not equal with nCat
			% extraLogZ_pos = zeros(1,nT);
			mixing = ones(1,nT) / nT;
			
			% apply prior on label uncertainty (this improves indeed)
			label_entropy = zeros(1,nT);
			for iT = 1:nT
				label_entropy(iT) = -sum( voting{iT} .* log(voting{iT}+1e-30) );
			end
			mixing2 = exp(-label_entropy * 5);
			mixing2 = mixing2/sum(mixing2)

        	% on training images:
        	for iCat = 1:nCat % for each category of images
        		load(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),'scores_train');
        		nimage = length(scores_train);
        		post_label_given_image_train = cell(nimage,1);
        		for i = 1:nimage % for each testing image
        			post_T_given_image_train = zeros( 1, nT );
        			for iT = 1:nT % for each template
        				post_T_given_image_train(iT) = exp(scores_train{i}(iT)-extraLogZ_pos(iT)) * mixing(iT) * mixing2(iT);
        			end
        			post_T_given_image_train = post_T_given_image_train / sum(post_T_given_image_train); % normalization

        			post_label_given_image_train{i} = zeros(1,nCat);
        			for iT = 1:nT
        				for jCat = 1:nCat
        					post_label_given_image_train{i}(jCat) = post_label_given_image_train{i}(jCat) +...
        						post_T_given_image_train(iT) * voting{iT}(jCat);
        				end
        			end
        		end
        		save(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),...
        			'-append','post_label_given_image_train');
        	end

        	% on testing images:
        	for iCat = 1:nCat % for each category of images
        		load(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),'scores');
        		nimage = length(scores);
        		post_label_given_image = cell(nimage,1);
        		for i = 1:nimage % for each testing image
        			post_T_given_image = zeros( 1, nT );
        			for iT = 1:nT % for each template
        				post_T_given_image(iT) = exp(scores{i}(iT)-extraLogZ_pos(iT)) * mixing(iT) * mixing2(iT);
        			end
        			post_T_given_image = post_T_given_image / sum(post_T_given_image); % normalization
        			
        			post_label_given_image{i} = zeros(1,nCat);
        			for iT = 1:nT
        				for jCat = 1:nCat
        					post_label_given_image{i}(jCat) = post_label_given_image{i}(jCat) + post_T_given_image(iT) * voting{iT}(jCat);
        				end
        			end
        		end
        		save(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),'-append','post_label_given_image');
        	end
        end

		
		%%% post processing by SVM
		function post_process_svm(this)
        	nCat = length(this.categories);
        	nT = nCat; % in future, nT may be not equal with nCat
			scaler = 1e-3;
			
        	% - on training images:
			% prepare dataMatrix
			dataMatrix = [];
			for iCat = 1:nCat
				load(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),'scores_train');
				nimage = length(scores_train);
				for i = 1:nimage
					dataMatrix = [dataMatrix;reshape(exp(scores_train{i}*scaler),1,nT)];
				end
			end

			svm_model = cell(nT,1);
        	for iCat = 1:nCat % for each category of images
				% train a 1-vs-all classifier by SVM
        		load(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),'scores_train');
        		nimage = length(scores_train);
				
				gt = zeros( size(dataMatrix,1), 1 );
				gt( (iCat-1) * length(scores_train) + (1:length(scores_train)) ) = 1;
				
				svm_model{iCat} = svmtrain(double(dataMatrix),gt,'Method','SMO');
				svm_model{iCat}.linearWeight = zeros(nT,1);
				tmp = svm_model{iCat}.SupportVectors;
				for indSV = 1:size(svm_model{iCat}.SupportVectors,1)
					svm_model{iCat}.linearWeight = svm_model{iCat}.linearWeight - svm_model{iCat}.Alpha(indSV) * ...
						reshape(svm_model{iCat}.SupportVectors(indSV,:),nT,1);
				end
        	end
			
			save(sprintf('Result/%s/GaborS1/voting_run%d.mat','summary',this.iRun), '-append', 'svm_model' );

			% on training images:
			for iCat = 1:nCat
				load(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),'scores_train');
				nimage = length(scores_train);
				post_label_given_image_train = cell(nimage,1);
				for i = 1:nimage
					vec = zeros(1,nT);
					datavec = zeros(nT,1);
					for iT = 1:nT
						datavec = exp(scores_train{i}(:)*scaler);
						if ~isempty(svm_model{iT}.ScaleData)
							datavec = svm_model{iT}.ScaleData.scaleFactor(:) .* ...
								( datavec(:) +  svm_model{iT}.ScaleData.shift(:));
						end
						vec(iT) = exp( ( sum( svm_model{iT}.linearWeight(:) .* datavec ) + svm_model{iT}.Bias ) );
					end
					svm_scores = log(vec);
					post_label_given_image_train{i} = vec / sum(vec);
				end

        		save(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),...
        			'-append','post_label_given_image_train','svm_scores');
			end

        	% on testing images:
        	for iCat = 1:nCat
				load(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),'scores');
				nimage = length(scores);
				post_label_given_image = cell(nimage,1);
				for i = 1:nimage
					vec = zeros(1,nT);
					datavec = zeros(nT,1);
					for iT = 1:nT
						datavec = exp(scores{i}(:)*scaler);
						if ~isempty(svm_model{iT}.ScaleData)
							datavec = svm_model{iT}.ScaleData.scaleFactor(:) .* ...
								( datavec(:) +  svm_model{iT}.ScaleData.shift(:));
						end
						vec(iT) = exp( sum( svm_model{iT}.linearWeight(:) .* datavec ) + svm_model{iT}.Bias );
					end
					post_label_given_image{i} = vec / sum(vec);
				end

        		save(sprintf('Result/%s/AdaptiveLocalLearning/out_run%d.mat', this.categories{iCat}, this.iRun),...
        			'-append','post_label_given_image');
			end
		end
		
    end
end




