classdef STF_MT % Sketch-Texture-Flat Multi-templates
    
    % ---------------------------------------------------------------------
    % Supported functionalities:
    %   
    %   1. Learning a large number of active basis/texture templates.
    %
    %   2. Cross-validation and testing category by category.
    %
    % 
    % Future work:
    %
    %   1. Use a standardized image pyramid and a better modularized
    %       convolution method.
    % ---------------------------------------------------------------------
    
    % member variables
    properties
        categories;	% collection of category names
        iRun;		% index for cross validation run
        iCategory;	% index of current category
        category;	% name of current category
        OverwriteModeIsOn = true;
		className = 'STF_MT';
		nPos = 15;
		fp_thres = 1/100;
    end

    properties
        maxImArea = 1000*1000;	% maximum allowed image area
        posFolder;		% folder for positive example images
        negFolder = 'Image/Natural';	% folder for random natural images
        featureVersion = 'GaborS1';		% feature name
        windowSize = 150; %
        Totalsketch = 60; % maximum allowed number of features
        epsilon = .1; % allowed correlation between selected Gabors
        sub = 1;    % subsampling rate for Gabor
        Lrange = 6; % allowed shifting in location (times sub pixels)
        Orange = 1; % allowed shifting in orientation
        norient = 16; % number of Gabor orientations
        %%% Note: Lrange*sub is the pixels allowed to shift along normal
        %%%       Lrange*sub should be the same in learning and testing
        SHUTUP = 0.; % SHUTUP is to neglect the positions with low average responses
        ToFindWhileLearning = true; % use EM-type iterations for learning templates for objects with unknown locations
        sizeGabor = 17;
		scale = 0.7;
        gainThres = 0.1;
		test_gainThres = 0.2;
        max_EM_iter = 20;
        displayModeIsOn = false;
        
        localHalfx = [];
		allsymbol = [];
		h = [];
		C = [];
		allfilter = [];
		
		NEGMAX = -1e10;
    end
    
    properties
        ap;				% average precision
        testAUC;		% area under ROC curve
        tp;				% true positive rates
        fp;				% false positive rates
        rec;			% recall rates
        prec;			% precision rates
        EER;			% equal-error rate
    end
    
    % functions
    methods
    
        % constructor
        function obj = STF_MT(iRun, iCat, categories)
            obj.iRun = iRun;
            obj.iCategory = iCat;
            obj.categories = categories;
            obj.category = categories{iCat};
        end
        
		function learn_all_category(this)
		
			% prepare Gabor filters and compute the inhibition map
            tic
            [allfilter, allsymbol] = makefilter(this.scale, this.norient);  % generate Gabor filters
            for o = 1:this.norient
                allsymbol{o} = single(allsymbol{o});
            end
            h = (size(allfilter{1}, 1)-1)/2;  % half size of Gabor
            C = corr(allfilter, this.epsilon);    % generate the inhibition maps
            for o = 1:this.norient
                for o2 = 1:this.norient
                    C{o,o2} = single(C{o,o2});
                end
            end
			this.allsymbol = allsymbol;
			this.h = h;
			this.C = C;
			this.allfilter = allfilter;
		
			for iter = 1:10
			
				% ---------------------------------
				% M step
				% ---------------------------------
				for iC = 1:length(this.categories)
					this.iCategory = iC; this.category = this.categories{iC};
					this.learn(iter);
				end
				
				% ---------------------------------
				% E step
				% ---------------------------------
				test_on_train_multiclass(this);
				assign_examples(this);
			end
		end

        %% learns multiple templates from a set of images (positive examples)
        function learn(this,iter)
            
            tic
            
            disp(sprintf('start learning: %d %s run %d, iteration %d',this.iCategory,this.category,this.iRun,iter));
            close all
            if this.displayModeIsOn
                figure;
                imshow( 255*ones(this.windowSize,this.windowSize),[] );
            end
            
			load('Feature/naturalStat_GaborS1_LocalNorm.mat'); % load natural statistics
			
            % TRAINING EXAMPLES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % ==================================================================
            %  input images from a folder
            % ==================================================================
            file0 = sprintf('Split/%s/split%d.mat',this.category,this.iRun);
            load(file0,'posTrainImages');
            nimage = length(posTrainImages);
			
			if iter == 1
				template_example_matrix = eye(nimage); % one example per template
			else
				load(sprintf('Result/%s/%s/out_run%d.mat',this.category,this.className,this.iRun),'template_example_matrix');
			end
            
            % load images
            stored_SUM1map = cell(nimage,this.norient);
			stored_SUM1LocalMean = cell(nimage,this.norient);
            for i = 1:nimage
                load(sprintf('Feature/%s/%s/%s.mat',...
                    this.featureVersion,this.category,posTrainImages{i}),...
                    'S1LocalNorm','RANGE');
				S1LocalNorm = convert2float(S1LocalNorm,RANGE);

				for o = 1:size(S1LocalNorm,2)
					tmp = inscribe([this.windowSize this.windowSize],S1LocalNorm{o},0);
					stored_SUM1map{i,o} = single(tmp);
				end

				load(sprintf('Feature/%s/%s/%s.mat',...
                    this.featureVersion,this.category,posTrainImages{i}),'S1LocalMean','localHalfx','RANGE');
				S1LocalMean = convert2float(S1LocalMean,RANGE);
				for o = 1:size(S1LocalMean,2)
					tmp = inscribe([this.windowSize this.windowSize],S1LocalMean{o},0);
					stored_SUM1LocalMean{i,o} = single(tmp);
				end
                
            end
			
			this.localHalfx = localHalfx;

			disp( sprintf('begin iteration %d', iter) );
			
			list_S2Template = cell(size(template_example_matrix,1),1);
			list_S2TemplateTexture = cell(size(template_example_matrix,1),1);
            if iter > 1
                filename = sprintf('Result/%s/%s/out_run%d.mat', this.category, this.className, this.iRun);
                load(filename,'row_trans','col_trans');
            end
            
            
			for iT = 1:size(template_example_matrix,1)

				ind = find(template_example_matrix(iT,:)); % top detections
				
                if iter == 1
                    row_translation = zeros(length(ind),1) + floor(this.windowSize/2);
                    col_translation = zeros(length(ind),1) + floor(this.windowSize/2);
                else
                    row_translation = zeros(length(ind),1);
                    for i = 1:length(row_translation)
                        if iT <= length(row_trans{ind(i)})
                            row_translation(i) = row_trans{ind(i)}(iT);
                            col_translation(i) = col_trans{ind(i)}(iT);
                        else
                            row_translation(i) = floor(this.windowSize/2);
                            col_translation(i) = floor(this.windowSize/2);
                        end
                    end
                end
                
                % crop the detected instances
                [SUM1mapLearn SUM1LocalMeanLearn] = ...
                    this.crop_instances( stored_SUM1map(ind,:),stored_SUM1LocalMean(ind,:), row_translation, col_translation );
				                
				%
				% learning of sketch
				%
				S2Template = this.learn_sketch( SUM1mapLearn );
				
				%
				% learning of texture
				%
				S2TemplateTexture = this.learn_tex( SUM1LocalMeanLearn );

				if this.displayModeIsOn
					imshow( S2Template.sym, [] );
					pause(0.2);
				end

				list_S2Template{iT} = S2Template;
				list_S2TemplateTexture{iT} = S2TemplateTexture;
			end
			
			save(sprintf('Result/%s/STF_MT/model_run%d.mat',this.category,this.iRun),...
				'list_S2Template','list_S2TemplateTexture','this');
            
            save(sprintf('Result/%s/STF_MT/model_run%d_iter%d.mat',this.category,this.iRun,iter),...
				'list_S2Template','list_S2TemplateTexture','this');

			disp(['learning time : ' num2str(toc) ' seconds, ' num2str(length(list_S2Template)) ' templates learned.']);
			
        end

        % -----------------------------------------------------
		%% test on training data
		% -----------------------------------------------------

		function test_on_train_multiclass(this)
        	nC = length(this.categories);
        	all_S2T_sketch = {};
        	all_S2T_texture = {};
        	templateIndPerCategory = cell(nC,1);
        	categoryPerTemplate = [];
        	% go over categories to fetch templates
        	offset = 0;
        	for iC = 1:nC
        		category = this.categories{iC};
        		load(sprintf('Result/%s/STF_MT/model_run%d.mat',category,this.iRun),'list_S2Template','list_S2TemplateTexture');
        		all_S2T_sketch = [ all_S2T_sketch; list_S2Template(:) ];
				all_S2T_texture = [ all_S2T_texture; list_S2TemplateTexture(:) ];
				templateIndPerCategory{iC} = offset + (1:length(list_S2Template));
				categoryPerTemplate = [categoryPerTemplate;ones(length(list_S2Template),1)*iC];
				offset = offset + length(list_S2Template);
        	end
        	
        	
        	for iT = 1:length(all_S2T_sketch)
        		S2Template = all_S2T_sketch{iT};
        		ind = find(S2Template.selectedGain > this.test_gainThres);
				S2Template.selectedRow = S2Template.selectedRow(ind);
				S2Template.selectedCol = S2Template.selectedCol(ind);
				S2Template.selectedOri = S2Template.selectedOri(ind);
				S2Template.selectedScale = S2Template.selectedScale(ind);
				S2Template.selectedLambda = S2Template.selectedLambda(ind);
				S2Template.selectedLogZ = S2Template.selectedLogZ(ind);
        		all_S2T_sketch{iT} = S2Template;

				S2TemplateTexture = all_S2T_texture{iT};
				ind = find(S2TemplateTexture.selectedGain > this.test_gainThres);
				S2TemplateTexture.selectedRow = S2TemplateTexture.selectedRow(ind);
				S2TemplateTexture.selectedCol = S2TemplateTexture.selectedCol(ind);
				S2TemplateTexture.selectedOri = S2TemplateTexture.selectedOri(ind);
				S2TemplateTexture.selectedScale = S2TemplateTexture.selectedScale(ind);
				S2TemplateTexture.selectedLambda = S2TemplateTexture.selectedLambda(ind);
				S2TemplateTexture.selectedLogZ = S2TemplateTexture.selectedLogZ(ind);
        		all_S2T_texture{iT} = S2TemplateTexture;
        	end
        	clear list_S2Template; clear list_S2TemplateTexture;
        	disp('templates loaded');
        	

        	% Go over categories to fetch images and compute SUM2 maps
        	count_image = 0;
        	for iCategory = 1:nC
        		fprintf(1,'testing on %s: ',this.categories{iCategory});
        		tic
		    	file0 = sprintf('Split/%s/split%d.mat',this.categories{iCategory},this.iRun);
		        load(file0,'posTrainImages');
		        nimage = length(posTrainImages);
		        count_image = count_image + nimage;
		        scores_train = cell(nimage,1);
				row_trans = cell(nimage,1);
				col_trans = cell(nimage,1);

		        for i = 1:nimage
		        	for kk = 1:length(num2str(i-1))
		        		if i == 1
		        		else
		        			fprintf('\b');
		        		end
		        	end
		        	fprintf(1,'%d',i);
				    % for each image, test all templates on it
				    load(sprintf('Feature/%s/%s/%s.mat',...
		                this.featureVersion,this.categories{iCategory},posTrainImages{i}),...
						'S1LocalMean','M1LocalNorm','RANGE');

					S1LocalMean = convert2float(S1LocalMean,RANGE);
					M1 = convert2float(M1LocalNorm,RANGE);

		            % - compute SUM2 maps
		            if ~exist( sprintf('Feature/GaborS2_%s/%s',this.className,this.categories{iCategory}), 'dir' )
						mkdir( sprintf('Feature/GaborS2_%s/%s',this.className,this.categories{iCategory}) );
					end
		            S2MapPath = sprintf('%s_%s/%s/%s_run%d.mat','Feature/GaborS2',this.className,this.categories{iCategory},posTrainImages{i},this.iRun);
					
					if ~exist( S2MapPath, 'file' ) || this.OverwriteModeIsOn
						subsampleS2 = 3;
						S2Map = mexc_ComputeSUM2( this.norient, M1, all_S2T_sketch, subsampleS2 );
						S2MapTexture = mexc_ComputeSUM2( this.norient, S1LocalMean, all_S2T_texture, subsampleS2 );
						
						for ii = 1:numel(S2Map)
							S2Map{ii} = S2Map{ii} + S2MapTexture{ii};
                            S2Map{ii} = imresize( S2Map{ii}, size(M1{1}), 'bilinear' );
                            
							S2Map{ii} = single(inscribe([this.windowSize this.windowSize],S2Map{ii},this.NEGMAX)); % inscribe the S2 map!
						end
						
						save(S2MapPath, 'S2Map');
					else % if the S2 map is already computed, then load it
						load(S2MapPath, 'S2Map');
                    end
                    
                    nT = length(all_S2T_sketch);
				    MAX2 = zeros(1,nT,'single');
					col_translation = zeros(1,nT,'single');
					row_translation = zeros(1,nT,'single');

				    for jT = 1:nT
				    	[MAX2(jT) bestInd] = max(S2Map{jT}(:));
						col_translation(jT) = ceil( bestInd / this.windowSize );
						row_translation(jT) = bestInd - (col_translation(jT)-1) * this.windowSize;
						col_translation(jT) = col_translation(jT) - 1;
						row_translation(jT) = row_translation(jT) - 1;
                    end
                    scores_train{i} = MAX2;
                    row_trans{i} = row_translation;
					col_trans{i} = col_translation;

                end

                filename = sprintf( 'Result/%s/STF_MT/out_run%d.mat', this.categories{iCategory}, this.iRun );
                if exist(filename,'file')
                    save(filename,'-append','scores_train','row_trans','col_trans','templateIndPerCategory','categoryPerTemplate');
                else
                    save(filename,'scores_train','row_trans','col_trans','templateIndPerCategory','categoryPerTemplate');
                end
		        fprintf(1,' images; takes %.1f sec\n',toc);
            end

            
            mixing = ones(1,nT) / nT;
			dif = 1;
			iIter = 1;
			
			last_voting = cell(1,nT);
			for iT = 1:nT
				last_voting{iT} = ones(1,nT) / nT;
			end

			while dif > 1e-6 && iIter < 2
				% now obtain voting statistics
				% voting{jT}{iCategory} means the conditional probability (not joint!) p( label | cluster number )
				voting = cell(1,nT);
				for iT = 1:nT
					voting{iT} = zeros( 1, length(this.categories) );
				end

				post_t_given_image = zeros(1,nT);
				new_mixing = zeros(1,nT) / nT;
				for iCategory = 1:length(this.categories)
					file0 = sprintf('Split/%s/split%d.mat',this.categories{iCategory},this.iRun);
					load(file0,'posTrainImages');
					nimage = length(posTrainImages);
					load(sprintf('Result/%s/STF_MT/out_run%d.mat', this.categories{iCategory}, this.iRun),'scores_train');

					for ii = 1:nimage
						for jT = 1:nT
							post_t_given_image(jT) = exp( scores_train{ii}(jT) ) * mixing(jT);
						end
						post_t_given_image = post_t_given_image / sum(post_t_given_image+1e-30);
						new_mixing = new_mixing + post_t_given_image;

						for jT = 1:nT
							voting{jT}(iCategory) = voting{jT}(iCategory) + post_t_given_image(jT);
						end
					end
				end

				new_mixing = new_mixing / count_image;
				dif = mean(abs(new_mixing-mixing));
				mixing = new_mixing;
				last_voting = voting;
				iIter = iIter + 1;
			end

			for iT = 1:nT
				voting{iT} = voting{iT} / sum(voting{iT}+1e-30);
			end

            mkdir(sprintf('Result/summary/%s',this.className));
		    save(sprintf('Result/%s/%s/voting_run%d.mat', 'summary', this.className, this.iRun),'voting','mixing','templateIndPerCategory','categoryPerTemplate');
        end

		%
		%% testing
        %
        function test_multiclass(this)
        	nT = length(this.categories);

        	% go over categories to fetch templates
        	list_of_S2Template = cell(nT,1);
        	for iT = 1:nT
        		category = this.categories{iT};
        		load(sprintf('Result/%s/STF_MT/model_run%d.mat',category,this.iRun),'S2Template','S2TemplateTexture','h');
        		ind = find(S2Template.selectedGain > this.test_gainThres);
				S2Template.selectedRow = S2Template.selectedRow(ind);
				S2Template.selectedCol = S2Template.selectedCol(ind);
				S2Template.selectedOri = S2Template.selectedOri(ind);
				S2Template.selectedScale = S2Template.selectedScale(ind);
				S2Template.selectedLambda = S2Template.selectedLambda(ind);
				S2Template.selectedLogZ = S2Template.selectedLogZ(ind);
        		list_of_S2Template{iT} = S2Template;

				ind = find(S2TemplateTexture.selectedGain > this.test_gainThres);
				S2TemplateTexture.selectedRow = S2TemplateTexture.selectedRow(ind);
				S2TemplateTexture.selectedCol = S2TemplateTexture.selectedCol(ind);
				S2TemplateTexture.selectedOri = S2TemplateTexture.selectedOri(ind);
				S2TemplateTexture.selectedScale = S2TemplateTexture.selectedScale(ind);
				S2TemplateTexture.selectedLambda = S2TemplateTexture.selectedLambda(ind);
				S2TemplateTexture.selectedLogZ = S2TemplateTexture.selectedLogZ(ind);
        		list_of_S2TemplateTexture{iT} = S2TemplateTexture;
        	end
        	disp('templates loaded');

        	% Go over categories to fetch images and compute SUM2 maps
        	for iCategory = 1:nT
        		fprintf(1,'testing on %s: ',this.categories{iCategory});
        		tic
		    	file0 = sprintf('Split/%s/split%d.mat',this.categories{iCategory},this.iRun);
		        load(file0,'posTestImages');
		        nimage = length(posTestImages);
		        scores = cell(nimage,1);
				row_trans = cell(nimage,1);
				col_trans = cell(nimage,1);

		        for i = 1:nimage
		        	for kk = 1:length(num2str(i-1))
		        		if i == 1
		        		else
		        			fprintf('\b');
		        		end
		        	end
		        	fprintf(1,'%d',i);
				    % for each image, test all templates on it
				    load(sprintf('Feature/%s/%s/%s.mat',...
		                this.featureVersion,this.categories{iCategory},posTestImages{i}),...
                        'S1LocalMean','M1LocalNorm','RANGE');
					S1LocalMean = convert2float(S1LocalMean,RANGE);
					M1 = convert2float(M1LocalNorm,RANGE);

		            % - compute SUM2 maps
		            if ~exist( sprintf('Feature/GaborS2/%s',this.categories{iCategory}), 'dir' )
						mkdir( sprintf('Feature/GaborS2/%s',this.categories{iCategory}) );
					end
		            S2MapPath = sprintf('%s/%s/%s_run%d.mat','Feature/GaborS2/',this.categories{iCategory},posTestImages{i},this.iRun);

					if ~exist( S2MapPath, 'file' ) || this.OverwriteModeIsOn
						subsampleS2 = 5;
						S2Map = mexc_ComputeSUM2( this.norient, M1, list_of_S2Template, subsampleS2 );
						S2MapTexture = mexc_ComputeSUM2( this.norient, S1LocalMean, list_of_S2TemplateTexture, subsampleS2 );
						for ii = 1:numel(S2Map)
							S2Map{ii} = S2Map{ii} + S2MapTexture{ii};
                            S2Map{ii} = imresize( S2Map{ii}, size(M1{1}) );
							S2Map{ii} = single(inscribe([this.windowSize this.windowSize],S2Map{ii},this.NEGMAX)); % inscribe the S2 map!
						end
						save(S2MapPath, 'S2Map');
					else % if the S2 map is already computed, then load it
						load(S2MapPath, 'S2Map');
					end

				    MAX2 = zeros(1,nT,'single');
				    col_translation = zeros(1,nT,'single');
					row_translation = zeros(1,nT,'single');
				    for jT = 1:nT
				    	[MAX2(jT) bestInd] = max(S2Map{jT}(:));
						col_translation(jT) = ceil( bestInd / this.windowSize );
						row_translation(jT) = bestInd - (col_translation(jT)-1) * this.windowSize;
						col_translation(jT) = col_translation(jT) - 1;
						row_translation(jT) = row_translation(jT) - 1;
                    end
                    scores_train{i} = MAX2;
                    row_trans{i} = row_translation;
					col_trans{i} = col_translation;
                end
                
                filename = sprintf('Result/%s/STF_MT/out_run%d.mat', this.categories{iCategory}, this.iRun);
                if exist(filename,'file')
                    save(filename,'-append','scores','row_trans','col_trans');
                else
                    save(filename,'scores','row_trans','col_trans');
                end
		        fprintf(1,' images; takes %.1f sec\n',toc);
            end
        end

        %%% post processing (Bayesian voting)

        function post_process(this)
        	load(sprintf('Result/%s/%s/voting_run%d.mat','summary',this.className,this.iRun), 'voting', 'extraLogZ_pos', 'mixing' );
        	nCat = length(this.categories);
        	nT = nCat; % in future, nT may be not equal with nCat
			mixing = ones(1,nT) / nT;
			
			% apply prior on label uncertainty (this improves indeed)
			label_entropy = zeros(1,nT);
			for iT = 1:nT
				label_entropy(iT) = -sum( voting{iT} .* log( voting{iT} + 1e-30 ) );
			end
			mixing2 = exp(-label_entropy * 0);
			mixing2 = mixing2/sum(mixing2+1e-10);
            extraLogZ_pos = zeros(1,nT);

        	% on training images:
        	for iCat = 1:nCat % for each category of images
        		load(sprintf('Result/%s/STF_MT/out_run%d.mat', this.categories{iCat}, this.iRun),'scores_train');
        		nimage = length(scores_train);
        		post_label_given_image_train = cell(nimage,1);
        		for i = 1:nimage % for each testing image
        			post_T_given_image_train = zeros( 1, nT );
        			for iT = 1:nT % for each template
        				post_T_given_image_train(iT) = exp(scores_train{i}(iT)-extraLogZ_pos(iT)) * mixing(iT) * mixing2(iT);
        			end
        			post_T_given_image_train = post_T_given_image_train / sum(post_T_given_image_train+1e-10); % normalization

        			post_label_given_image_train{i} = zeros(1,nCat);
        			for iT = 1:nT
        				for jCat = 1:nCat
        					post_label_given_image_train{i}(jCat) = post_label_given_image_train{i}(jCat) +...
        						post_T_given_image_train(iT) * voting{iT}(jCat);
        				end
        			end
        		end
        		save(sprintf('Result/%s/STF_MT/out_run%d.mat', this.categories{iCat}, this.iRun),...
        			'-append','post_label_given_image_train');
        	end

        	% on testing images:
        	for iCat = 1:nCat % for each category of images
        		load(sprintf('Result/%s/STF_MT/out_run%d.mat', this.categories{iCat}, this.iRun),'scores');
        		nimage = length(scores);
        		post_label_given_image = cell(nimage,1);
        		for i = 1:nimage % for each testing image
        			post_T_given_image = zeros( 1, nT );
        			for iT = 1:nT % for each template
        				post_T_given_image(iT) = exp(scores{i}(iT)-extraLogZ_pos(iT)) * mixing(iT) * mixing2(iT);
        			end
        			post_T_given_image = post_T_given_image / sum(post_T_given_image+1e-10); % normalization
        			
        			post_label_given_image{i} = zeros(1,nCat);
        			for iT = 1:nT
        				for jCat = 1:nCat
        					post_label_given_image{i}(jCat) = post_label_given_image{i}(jCat) + post_T_given_image(iT) * voting{iT}(jCat);
        				end
        			end
        		end
        		save(sprintf('Result/%s/STF_MT/out_run%d.mat', this.categories{iCat}, this.iRun),'-append','post_label_given_image');
        	end
        end

		
		function S2Template = learn_sketch( this, SUM1mapLearn )
	
			load('Feature/naturalStat_GaborS1_LocalNorm.mat'); % load natural statistics
			
			allsymbol = this.allsymbol;
			h = this.h;
			C = this.C;
            
            nimage = size(SUM1mapLearn,1);
			
			Mi = zeros(1, this.Totalsketch, 'single');  % orientation of selected Gabor
			Mx = zeros(1, this.Totalsketch, 'single');  % position of selected Gabor
			My = zeros(1, this.Totalsketch, 'single'); 
			Mm = zeros(1, this.Totalsketch, 'single');  % lambda
			Mm1 = zeros(1, this.Totalsketch, 'single');  % logZ
			gain = zeros(1, this.Totalsketch, 'single'); % information gains (bits)
			sym = zeros(this.windowSize,this.windowSize, 'single');  % symbolic plot of selected Gabor
			Asym = cell(1, nimage);   % symbolic plot for each image with active Gabor
			for i = 1 : nimage
				Asym{i} = zeros(size(SUM1mapLearn{i,1}),'single');  
			end
			tss = zeros(nimage, 1, 'single');  % total scores for testing images
			gain = zeros(this.Totalsketch, 1, 'single');  % coding gain for selected Gabor
			binnum = 50; % number of bins
			subsampleM1 = 1;
			MAX1mapLearn = cell(size(SUM1mapLearn));
			ARGMAX1mapLearn = cell(size(SUM1mapLearn));
			for i = 1:nimage
				[MAX1mapLearn(i,:) ARGMAX1mapLearn(i,:) M1RowShift M1ColShift M1OriShifted] = ...
					mexc_ComputeMAX1( this.norient, SUM1mapLearn(i,:), 6.0/17.0,...
						this.Orange, single(h*2+1), subsampleM1 );
			end
			nimage = size(SUM1mapLearn,1);
			mexc_SharedSketch(this.norient, this.Lrange * this.sub, this.Orange, subsampleM1, ... % about active basis  
			   this.Totalsketch, nimage, this.windowSize, this.windowSize, ...
			   SUM1mapLearn, MAX1mapLearn, ARGMAX1mapLearn, ... % about training images 
			   h, C, allsymbol(1, :), ... % about filters
			   single(M), single(lam), single(e), single(lz), ... % about exponential model 
			   Mi, Mx, My, Mm, Mm1,gain, ... % learned parameters
			   sym, Asym,... % learned templates 
			   M1RowShift, M1ColShift, M1OriShifted);
			
			ind = find(gain>this.gainThres);
			Mi = Mi(ind);
			Mx = Mx(ind);
			My = My(ind);
			Mm = Mm(ind);
			Mm1 = Mm1(ind);
			gain = gain(ind);
			sym = displayGaborTemplate([this.windowSize this.windowSize], Mx,My,Mi,Mm);
			
			if ~exist(sprintf('Result/%s/STF_MT/',this.category),'dir')
				mkdir(sprintf('Result/%s/STF_MT/',this.category));
			end
			
			Lrange = this.Lrange; Orange = this.Orange; sub = this.sub;
			S2Template = struct( 'selectedRow', single(Mx)-1-floor(this.windowSize/2), ...
				'selectedCol', single(My)-1-floor(this.windowSize/2), ...
				'selectedOri', single(Mi), ...
				'selectedScale', zeros(1,length(Mx),'single'), ... 
				'selectedLambda', single(Mm), ...
				'selectedLogZ', single(Mm1), ...
				'subsampleM1', 1, ...
				'selectedGain', gain,...
				'sym', sym ...
				);
			
			
		end
		
		
		function S2TemplateTexture = learn_tex( this, SUM1LocalMeanLearn )

			load('Feature/naturalStat_GaborS1_LocalMean.mat'); % load natural statistics
			storedMean = e;
			storedLambda = lam;
			storedLogZ = lz;
			localHalfx = this.localHalfx;
            h = this.h;
			templateSize = size(SUM1LocalMeanLearn{1});
			nimage = size(SUM1LocalMeanLearn,1);
			meanLocalMeanMap = zeros(size(SUM1LocalMeanLearn{1},1), size(SUM1LocalMeanLearn{1},2),length(SUM1LocalMeanLearn));
			for i = 1:nimage
				for o = 1:norient
					meanLocalMeanMap(:,:,o) = meanLocalMeanMap(:,:,o) + 1/nimage * SUM1LocalMeanLearn{i,o};
				end
			end
			
			maxNumFeature = this.Totalsketch;
			Mi = zeros(1, maxNumFeature);  % orientation of selected Gabor
			Mx = zeros(1, maxNumFeature);  % position of selected Gabor
			My = zeros(1, maxNumFeature); 
			Mm = zeros(1, maxNumFeature);  % mean
			lambda = zeros(1, maxNumFeature);  % lambda
			logZ = zeros(1, maxNumFeature);  % logZ
			gain = zeros(maxNumFeature, 1);  % coding gain for selected Gabor

			gainMap = zeros(templateSize(1),templateSize(2),this.norient);
			ind = zeros(templateSize(1),templateSize(2),o);
			ax = localHalfx;
			for x = (h+ax)+1:templateSize(1)-(h+ax)
				for y = 1+(h+ax):templateSize(2)-(h+ax)
					for o = 1:this.norient
						[dif ind(x,y,o)] = min( abs(storedMean - meanLocalMeanMap(x,y,o)) );
						ind(x,y,o) = min(length(storedMean),max(1,ind(x,y,o)));
						gainMap(x,y,o) = storedMean(ind(x,y,o)) * storedLambda(ind(x,y,o)) - storedLogZ(ind(x,y,o));
					end
				end
			end
			inhibited = zeros(templateSize(1),templateSize(2),this.norient);

			nF = maxNumFeature;
			for k = 1:nF
				[maxGain here] = max(gainMap(:) .* double(1-inhibited(:))  );
				binNo = ind(here);

				if maxGain <= this.gainThres
					nF = k-1;
					break
				end

				ori = ceil(here/templateSize(1)/templateSize(2)); % start from 1 !!
				Mi(k) = ori;
				col = ceil((here - (ori-1) * templateSize(1) * templateSize(2))/templateSize(1));
				My(k) = col;
				row = here - (ori-1) * templateSize(1) * templateSize(2) - (col-1)*templateSize(1); 
				Mx(k) = row;
				
				gain(k) = maxGain;
				Mm(k) = storedMean(binNo);
				lambda(k) = storedLambda(binNo);
				logZ(k) = storedLogZ(binNo);
				
				% do inhibition
				left = max(1,col-ax-h);
				right = min(templateSize(2),col+ax+h);
				top = max(1,row-ax-h);
				bottom = min(templateSize(1),row+ax+h);
				for iOri = ori-1:ori+1
					iOri2 = iOri;
					if iOri < 1
						iOri2 = iOri + this.norient;
					end
					if iOri > this.norient
						iOri2 = iOri - this.norient;
					end
					inhibited(top:bottom,left:right,iOri2) = 1;
					gainMap(top:bottom,left:right,iOri2) = 0;
				end
			end
			Mm = Mm(1:nF); Mx = Mx(1:nF); My = My(1:nF);
			gain = gain(1:nF); Mi = Mi(1:nF);
			lambda = lambda(1:nF);
			logZ = logZ(1:nF);
			
			S2TemplateTexture = struct( 'selectedRow', single(Mx)-1-floor(this.windowSize/2), ...
				'selectedCol', single(My)-1-floor(this.windowSize/2), ...
				'selectedOri', single(Mi)-1, ...
				'selectedScale', zeros(1,length(Mx),'single'), ... 
				'selectedLambda', single(lambda), ...
				'selectedLogZ', single(logZ), ...
				'subsampleM1', 1, ...
				'selectedGain', gain,...
				'sym', sym ...
				);
			

		end
		
		
		
		function [SUM1mapLearn SUM1LocalMeanLearn] ...
				= crop_instances( this, stored_SUM1map,stored_SUM1LocalMean, row_translation, col_translation )
			% - crop SUM1 (sigmoid saturated) maps
			denseX = -floor(this.windowSize/2) + (1:this.windowSize)-1;
			denseY = denseX;
			count = 0;
			inRow = zeros(length(denseX)*length(denseY),1,'single');
			inCol = zeros(length(denseX)*length(denseY),1,'single');
			inOri = zeros(length(denseX)*length(denseY),1,'single');
			inScale = zeros(length(denseX)*length(denseY),1,'single');
			for y = denseY
				for x = denseX
					count = count+1;
					inRow(count) = x;
					inCol(count) = y;
				end
			end
			
			SUM1mapLearn = cell(length(row_translation),this.norient);
			SUM1LocalMeanLearn = cell(length(row_translation),this.norient);
			
			for iImage = 1:length(row_translation)

				% use mex-C code to transform the dense grid, which is used to fetch SUM1 map pixels
				[outX outY outOri] = mexc_TemplateAffineTransform( 0, 2^0, 2^0, 0, inRow, inCol, inOri, inScale, this.norient );
				
				tScale = 0; destHeight = this.windowSize; destWidth = this.windowSize; nScale = 1;
				SUM1mapLearn(iImage,:) = mexc_CropInstance( stored_SUM1map(iImage,:),...
					row_translation(iImage),...
					col_translation(iImage),...
					0,tScale,outX,outY,destWidth,destHeight,this.norient,1 );

				SUM1LocalMeanLearn(iImage,:) = mexc_CropInstance( stored_SUM1LocalMean(iImage,:),...
					row_translation(iImage),...
					col_translation(iImage),...
					0,tScale,outX,outY,destWidth,destHeight,this.norient,1 );
					
			end
		end
		
		function assign_examples(this)
			% load the score matrix
			tssMat = [];
			for iCat = 1:length(this.categories)
				load(sprintf('Result/%s/%s/out_run%d.mat', this.categories{iCat}, this.className, this.iRun),'scores_train',...
					'templateIndPerCategory', 'categoryPerTemplate');
				for i = 1:length(scores_train)
					tssMat = [tssMat; scores_train{i}(1,:)];
				end
			end
			numImage = size(tssMat,1);
			numTemplate = size(tssMat,2);
			thres = zeros(numTemplate,1);
			
			
			% obtain score matrix and thresholds per template
			for iT = 1:numTemplate
                iCat = categoryPerTemplate(iT);
				% only read the scores on other categories
				ind = [ max(1,(iCat-2)*this.nPos+1):max(0,(iCat-1)*this.nPos) ...
					min(numImage+1, iCat*this.nPos+1) : min(numImage, (iCat+1)*this.nPos) ];
				scores_per_T = tssMat(ind,iT);
				[sorted idx] = sort(scores_per_T,'descend');
				cut_off = max(1,floor(this.fp_thres * numImage));
				thres(iT) = sorted(cut_off);
			end


			% cluster pursuit
			for iCat = 1:length(this.categories)
				% extract the examples and templates associated with this category
				
				partialMat = tssMat( (iCat-1)*this.nPos+(1:this.nPos), templateIndPerCategory{iCat} );
				partialThres = thres( templateIndPerCategory{iCat} );
				[template_example_matrix selectedTemplateNo] = this.cluster_pursuit(partialMat, partialThres);
				save(sprintf('Result/%s/%s/out_run%d.mat', this.categories{iCat}, this.className, this.iRun),...
					'-append','template_example_matrix','selectedTemplateNo');
				disp( sprintf('Result/%s/%s/out_run%d.mat', this.categories{iCat}, this.className, this.iRun) );
			end
			
		end
		
		% select a compact subset of templates that "cover" the positive examples
		function [template_example_matrix selectedTemplateNo] = cluster_pursuit(this, scores, thres)
			nimage = size(scores,1);
			nT = size(scores,2); % total number of templates
			gains = zeros( nT, 1 );
			
            
			selectedTemplateNo = [];
			template_example_matrix = [];
			covered = zeros(nimage,1);
			for iSelected = 1:nT
				
				for jT = 1:nT
					max_pos = max( scores(covered==0,jT) ) - 1e-6; % the maximum score of uncovered examples
					thres(jT) = min( max_pos, thres(jT) );
				end
				
				on_off = zeros( nimage, nT );
				for iT = 1:nT
					on_off(:,iT) = scores(:,iT) >= thres(iT);
				end
				on_off(covered==1,:) = 0;
			
				% update gains
				gains = sum(on_off,1);
                gains(selectedTemplateNo) = -1;

				% select
				[maxgain templateNo] = max(gains);
				selectedTemplateNo = [selectedTemplateNo templateNo];
				
				template_example_matrix = [template_example_matrix on_off(:,templateNo)];

				% inhibition
				ind = find(on_off(:,templateNo));
                covered(ind) = 1;

				if sum(covered==0) == 0 % all examples are covered
					break;
				end
            end
			
            tmp = sum(template_example_matrix,2);
            if sum(covered(:) ~= tmp(:)) > 0
                disp('wrong!');
                disp(covered);
                disp(template_example_matrix);
                pause
            end
            
			% there could be left-over examples unexplained!

            unexplained = find(covered==0);
            if ~isempty(unexplained)
                for i = unexplained'
                    if i > 0 || i < length(covered)
                        toadd = zeros( size(template_example_matrix,1), 1 );
                        toadd(i) = 1;
                        template_example_matrix = [template_example_matrix toadd];
                        % disp('added');
                    end
                end
            end
            
			
			if sum(sum(template_example_matrix,2) == 1) ~= size(template_example_matrix,1)
				disp('wrong here');
				disp(template_example_matrix);
				disp(covered);
                disp(gains);
                disp(selectedTemplateNo);
				pause
			end
			
			template_example_matrix = template_example_matrix';
			
		end
		
    end
end




